#pragma once

#include "apply.hpp"
#include "types.hpp"

class TileLoaderBase
{
public:
  static constexpr int     BLOCK_SIZE     = 128;
  virtual image2d<uint8_t> load(rect roi) = 0;
};

class TileLoaderCache : public TileLoaderBase
{
public:
  TileLoaderCache();
  virtual ~TileLoaderCache();

  TileLoaderCache(const TileLoaderCache&) = delete;
  TileLoaderCache& operator=(const TileLoaderCache&) = delete;
  TileLoaderCache(TileLoaderCache&&)                 = delete;
  TileLoaderCache& operator=(TileLoaderCache&&) = delete;

protected:
  image2d<uint8_t> m_tile;
};

template <class I>
class TileLoader : public TileLoaderCache
{
public:
  TileLoader(I input)
    : m_input(input)
  {
  }

  image2d<uint8_t> load(rect roi) final
  {
    // Because of the border
    roi.x -= 1;
    roi.y -= 1;
    roi.width += 2;
    roi.height += 2;
    copy_to(m_input, m_tile, roi);
    return m_tile.shift({1, 1});
  }

private:
  I m_input;
};

template <>
class TileLoader<image2d<uint8_t>> : public TileLoaderBase
{
public:
  TileLoader(image2d<uint8_t> input)
    : m_input(input)
  {
  }

  image2d<uint8_t> load(rect roi) final { return m_input.shift({roi.x, roi.y}); }

private:
  image2d<uint8_t> m_input;
};


namespace details
{
  void dilate8(TileLoaderBase* loader, image2d<uint8_t> out, rect roi);
}

template <class I>
void dilate8(I input, image2d<uint8_t> out, rect roi)
{
  TileLoader<I> loader(input);
  details::dilate8(&loader, out, roi);
}
