#pragma once

#include <cstddef>
#include <cstdint>
#include <type_traits>

struct rect
{
  int x;
  int y;
  int width;
  int height;
};


struct point
{
  int x;
  int y;
};


struct rgb16
{
  uint16_t r, g, b;
};

template <class T>
struct image2d;

template <>
struct image2d<void>
{
  std::byte*     m_buffer;
  std::ptrdiff_t m_byte_stride;
};


template <class I>
concept Image2D = std::is_invocable_v<I, point>;


template <class T>
struct image2d : image2d<void>
{
  T& operator()(point p) const
  {
    T* lineptr = (T*)((char*)m_buffer + m_byte_stride * p.y);
    return lineptr[p.x];
  }

  image2d<T> shift(point p) const
  {
    image2d<T> tmp;
    tmp.m_buffer      = m_buffer + m_byte_stride * p.y + sizeof(T) * p.x;
    tmp.m_byte_stride = m_byte_stride;
    return tmp;
  }

  T* buffer() const { return (T*)m_buffer; }

  std::ptrdiff_t stride() const { return m_byte_stride / sizeof(T); }
};
