#include "dilate.hpp"

#include <algorithm>

namespace details
{
  namespace
  {
    void dilate8(image2d<uint8_t> input, image2d<uint8_t> output, int width, int height)
    {
      uint8_t* __restrict A = input.buffer();
      uint8_t* __restrict B = output.buffer();

      for (int y = 0; y < height; ++y)
      {
        auto a = A - input.stride();
        auto b = A;
        auto c = A + input.stride();

        for (int x = 0; x < width; ++x)
        {
          int m = b[x];
          if (a[x - 1] > m)
            m = a[x - 1];
          if (a[x] > m)
            m = a[x];
          if (a[x + 1] > m)
            m = a[x + 1];
          if (b[x - 1] > m)
            m = b[x - 1];
          // if (b[x]   > m)   m = b[x];
          if (b[x + 1] > m)
            m = b[x + 1];
          if (c[x - 1] > m)
            m = c[x - 1];
          if (c[x] > m)
            m = c[x];
          if (c[x + 1] > m)
            m = c[x + 1];
          B[x] = m;
        }
        A += input.stride();
        B += output.stride();
      }
    }
  } // namespace


  void dilate8(TileLoaderBase* loader, image2d<uint8_t> out, rect roi)
  {
    for (int y = 0; y < roi.height; y += TileLoaderBase::BLOCK_SIZE)
      for (int x = 0; x < roi.width; x += TileLoaderBase::BLOCK_SIZE)
      {
        int w = std::min(roi.width - x, TileLoaderBase::BLOCK_SIZE);
        int h = std::min(roi.height - y, TileLoaderBase::BLOCK_SIZE);

        rect r = {roi.x + x, roi.y + y, w, h};

        auto tile_in  = loader->load(r);
        auto tile_out = out.shift({x, y});
        dilate8(tile_in, tile_out, w, h);
      }
  }
} // namespace details


TileLoaderCache::TileLoaderCache()
{
  int sz               = TileLoaderBase::BLOCK_SIZE + 2;
  m_tile.m_buffer      = (std::byte*)std::malloc(sz * sz * sizeof(uint8_t));
  m_tile.m_byte_stride = sz;
}


TileLoaderCache::~TileLoaderCache()
{
  std::free(m_tile.m_buffer);
}
