#include "algorithms.hpp"
#include "dilate.hpp"
#include "image.hpp"

#include <cassert>
#include <chrono>
#include <iostream>

void chain1(image2d<rgb16> input, image2d<uint8_t> output, rect roi)
{
  auto A = create_image<uint16_t>(roi.width, roi.height);
  auto B = create_image<uint8_t>(roi.width, roi.height);

  togray(input, A, roi);
  subquantize8(A, B, roi);

  roi.x += 1;
  roi.y += 1;
  roi.width -= 2;
  roi.height -= 2;
  dilate8(B, output, roi);

  destroy_image(A);
  destroy_image(B);
}

void chain2(image2d<rgb16> input, image2d<uint8_t> output, rect roi)
{
  auto A = togray_view(input);
  auto B = subquantize8_view(A);

  roi.x += 1;
  roi.y += 1;
  roi.width -= 2;
  roi.height -= 2;
  dilate8(B, output, roi);
}

void test(image2d<rgb16> input, rect roi, int width, int height)
{
  std::cout << "Testing output...";

  auto output1 = create_image<uint8_t>(width, height);
  chain1(input, output1, roi);

  auto output2 = create_image<uint8_t>(width, height);
  chain1(input, output2, roi);

  bool same = true;
  for (int x = 0; x < width; ++x)
    for (int y = 0; y < height; ++y)
      same &= output1({x, y}) == output2({x, y});

  assert(same);
  std::cout << " OK" << std::endl;
}


int main()
{
  int  width  = 5160;
  int  height = 3872;
  rect roi    = {0, 0, width, height};

  auto input  = create_image<rgb16>(width, height);
  auto output = create_image<uint8_t>(width, height);
  fill_with_random_values(input, roi);

#ifdef TESTING
  test(input, roi, width, height);
#endif

  std::chrono::duration<double> d1, d2;
  {
    auto start = std::chrono::system_clock::now();
    chain1(input, output, roi);
    auto end = std::chrono::system_clock::now();
    d1       = end - start;
  }
  {
    auto start = std::chrono::system_clock::now();
    chain2(input, output, roi);
    auto end = std::chrono::system_clock::now();
    d2       = end - start;
  }
  destroy_image(input);
  destroy_image(output);

  std::cout << "Normal pipeline: " << d1.count() << "s\n"
            << "View pipeline: " << d2.count() << "s\n";
}
